#!/usr/bin/env python
from __future__ import division
import time
from math import atan2, pi
from threading import Thread

from screen import screen_size
from tuio_server import TuioServer2D
from logger import Logger
from events import Event, DownEvent, UpEvent, MoveEvent, Tap, SingleTap, \
        DoubleTap, Rotate, Pinch, Pan

# Heuristic constants
# TODO: Encapsulate screen resolution in distance heuristics
SUPPORTED_EVENTS = ('down', 'up', 'move', 'tap', 'single_tap', 'double_tap',
                    'pan', 'flick', 'rotate', 'pinch')
DOUBLE_TAP_DISTANCE = .05
FLICK_VELOCITY_TRESHOLD = 20
TAP_TIMEOUT = .2
MAX_MULTI_DRAG_DISTANCE = .05
STATIONARY_TIME = .01

# Detect gestures 60 times per second
GESTURE_UPDATE_RATE = 60

# Minimum distance for two coordinates to be considered different
# Theoretically, this should be one pixel because that is the minimal movement
# of a mouse cursor on the screen
DIST_THRESHOLD = 1. / max(screen_size)


def rotation(p0, p1):
    """
    Calculate the relative rotation from p0 to p1 around (0, 0) in radians.
    """
    rot = atan2(*p1) - atan2(*p0)

    # Assert that the smallest possible rotation value is returned
    if rot >= pi:
        rot = 2 * pi - rot
    elif rot <= -pi:
        rot = -2 * pi - rot

    return rot


def distance(a, b):
    """
    Calculate the Pythagorian distance between points a and b (which are (x, y)
    tuples).
    """
    xa, ya = a
    xb, yb = b

    return ((xb - xa) ** 2 + (yb - ya) ** 2) ** .5


def average(points):
    l = len(points)
    x, y = zip(*points)

    return sum(x) / l, sum(y) / l


# Maximum distance between touch- and release location of a tap event
TAP_DISTANCE = .01

# Maximum duration of a tap
TAP_TIME = .300


class TouchPoint(object):
    def __init__(self, sid, x, y):
        self.start_time = self.update_time = time.time()
        self.sid = sid
        self.start_x = self.px = self.x = x
        self.start_y = self.py = self.y = y
        self.prev_angle = None

    @property
    def xy(self):
        return self.x, self.y

    def dx(self):
        return self.x - self.px

    def dy(self):
        return self.y - self.py

    def update(self, x, y):
        self.update_time = time.time()
        self.px = self.x
        self.py = self.y
        self.x = x
        self.y = y

    def down_time(self):
        return time.time() - self.start_time

    def angle(self, cx, cy):
        return atan2(cy - self.y, self.x - cx)

    def distance(self, other_x, other_y):
        return distance((self.x, self.y), (other_x, other_y))

    def is_tap(self):
        return time.time() - self.update_time < TAP_TIME \
               and self.distance(self.start_x, self.start_y) < TAP_DISTANCE

    def distance_to_prev(self):
        return self.distance(self.px, self.py)

    def is_stationary(self):
        return self.distance_to_prev() < DIST_THRESHOLD

    def rotation_around(self, cx, cy):
        prev = self.px - cx, self.py - cy
        current = self.x - cx, self.y - cy

        return rotation(prev, current)


class MultitouchServer(Logger):
    def __init__(self, verbose=0, tuio_verbose=0, **kwargs):
        super(MultitouchServer, self).__init__(**kwargs)
        self.verbose = verbose
        self.last_tap = None
        self.last_tap_time = 0
        self.handlers = {}
        self.thread = None
        self.points_changed = False

        # Session id's pointing to point coordinates
        self.points = []

        # Put centroid outside screen to prevent misinterpretation
        self.centroid = -1., -1.
        self.rot_centroid = -1., -1.

        self.server = TuioServer2D(self, verbose=tuio_verbose)

        self.angle = self.pinch = None
        self.prev_dist = None

    def point_down(self, sid, x, y):
        """
        Called by TUIO server when a new touch point is created, triggers a
        DownEvent.
        """
        if self.find_point(sid):
            raise ValueError('Point with session id "%d" already exists.'
                             % sid)

        p = TouchPoint(sid, x, y)
        self.points.append(p)
        self.trigger(DownEvent(p))
        self.points_changed = True

    def point_up(self, sid):
        """
        Called by TUIO server when a touch point is removed, triggers an
        UpEvent. Also, simple/double tap detection is located here instead of
        in the gesture thread (for responsiveness reasons).
        """
        i, p = self.find_point(sid, index=True)

        if not p:
            raise KeyError('No point with session id "%d".' % sid)

        del self.points[i]
        self.trigger(UpEvent(p))

        if p.is_tap():
            # Always trigger a regular tap event, also in case of double tap
            # (use the 'single_tap' event to keep single/double apart from
            # eachother)
            self.trigger(Tap(p.x, p.y))

            # Detect double tap by comparing time and distance to last tap
            # event
            t = time.time()

            if t - self.last_tap_time < TAP_TIMEOUT \
                    and p.distance(*self.last_tap.xy) < DOUBLE_TAP_DISTANCE:
                self.trigger(DoubleTap(p.x, p.y))
                self.last_tap = None
                self.last_tap_time = 0
            else:
                self.last_tap = p
                self.last_tap_time = t

        # TODO: Detect flick
        #elif p.is_flick():
        #    self.trigger(Flick(p.x, p.y))

        self.points_changed = True

    def point_move(self, sid, x, y):
        """
        Called by TUIO server when a touch point moves, triggers a MoveEvent.
        The move event is only used if the movement distance is greater that a
        preset constant, so that negligible movement is ignored. This prevents
        unnecessary gesture detection.
        """
        p = self.find_point(sid)

        if p.distance(x, y) > DIST_THRESHOLD:
            p.update(x, y)
            self.trigger(MoveEvent(p))
            self.points_changed = True

    def detect_rotation_and_pinch(self):
        """
        Rotation is the average angle change between each point and the
        centroid. Pinch is the percentual change in average distance change
        from the points to the centroid.
        """
        l = len(self.points)

        if l < 2 or ('pinch' not in self.handlers \
                     and 'rotate' not in self.handlers):
            self.prev_dist = None
            return

        l = len(self.points)
        cx, cy = self.centroid
        rotation = sum([p.rotation_around(cx, cy) for p in self.points]) / l
        dist = sum([p.distance(cx, cy) for p in self.points]) / l

        if rotation:
            self.trigger(Rotate(cx, cy, rotation, l))

        if self.prev_dist != None:
            pinch = dist / self.prev_dist

            if pinch:
                self.trigger(Pinch(cx, cy, pinch, l))

        self.prev_dist = dist

    def update_centroid(self):
        """
        Calculate the centroid of all current touch points.
        """
        self.old_centroid = self.centroid

        if self.points:
            self.centroid = average([p.xy for p in self.points])
        else:
            # There are no touch points, move the centroid out of the screen
            self.centroid = -1., -1.

    def centroid_movement(self):
        cx, cy = self.centroid
        ocx, ocy = self.old_centroid

        return cx - ocx, cy - ocy

    #def detect_pan(self):
    #    """
    #    Look for multi-finger drag events. Multi-drag is defined as all the
    #    fingers moving close-ish together in the same direction.
    #    """
    #    l = len(self.points)

    #    if not l:
    #        return False

    #    m = MAX_MULTI_DRAG_DISTANCE
    #    clustered = l == 1 or all([p.distance(*self.centroid) <= m \
    #                               for p in self.points])
    #    directions = [(cmp(p.dx(), 0), cmp(p.dy(), 0))
    #                  for p in self.points]

    #    if not clustered or not any(map(all, zip(*directions))):
    #        return False

    #    if l == 1:
    #        p = self.points[0]
    #        cx, cy, dx, dy = p.x, p.y, p.dx(), p.dy()
    #    else:
    #        cx, cy = self.centroid
    #        old_cx, old_cy = self.old_centroid
    #        dx, dy = cx - old_cx, cy - old_cy

    #    self.trigger(Pan(cx, cy, dx, dy, l))

    #    return True

    def detect_pan(self):
        """
        Panning is movement of the centroid in a certain direction.
        """
        l = len(self.points)

        if not l:
            return

        if distance(self.centroid, self.old_centroid) > DIST_THRESHOLD:
            cx, cy = self.centroid
            dx, dy = self.centroid_movement()
            self.trigger(Pan(cx, cy, dx, dy, l))

    def find_point(self, sid, index=False):
        for i, p in enumerate(self.points):
            if p.sid == sid:
                return (i, p) if index else p

        if index:
            return -1, None

    def detect_single_tap(self):
        """
        Check if a single tap event should be triggered by checking is the last
        tap.
        """
        if self.last_tap and time.time() - self.last_tap_time >= TAP_TIMEOUT:
            self.trigger(SingleTap(*self.last_tap.xy))
            self.last_tap = None
            self.last_tap_time = 0

    def detect_gestures(self):
        """
        Detect if any gestures have occured in the past gesture frame. This
        method is called in each time interval of the gesture thread, for
        gestures that can only be detected using accumulated point down/up/move
        events.
        """
        # Simple and double taps are detected in the main thread, specific
        # single-tap in the gesture thread
        self.detect_single_tap()

        # If the touch points have not been updated, neither have the gestures
        if self.points_changed:
            # The new centroid is used for pan/rotate/pinch detection
            self.update_centroid()

            # If a pan event is detected, ignore any rotate or pinch movement
            # (they are considered noise)
            self.detect_pan()
            self.detect_rotation_and_pinch()
            #if not self.detect_pan():
            #    self.detect_rotation_and_pinch()

            self.points_changed = False

    def start_gesture_thread(self):
        """
        Loop of the gesture thread.
        """
        interval = 1. / GESTURE_UPDATE_RATE

        while True:
            self.detect_gestures()
            time.sleep(interval)

    def stop(self):
        """
        Stop main event loop.
        """
        self.log('Stopping TUIO server')
        self.server.stop()

        if self.thread:
            self.log('Stopping main loop thread')
            self.thread.join()
            self.thread = None

    def start(self, threaded=False, daemon=False):
        """
        Start main event loop. If threaded is set to True, the main loop is
        started in a new thread. If daemon is also set to True, that thread
        will be daemonic. The daemon option makes a call to stop() unnecessary.
        """
        if threaded:
            self.log('Creating %sthread for main loop'
                     % ('daemon ' if daemon else ''))
            self.thread = Thread(target=self.start)
            self.thread.daemon = daemon
            self.thread.start()
            return

        # Start gesture thread
        self.log('Starting gesture thread')
        gesture_thread = Thread(target=self.start_gesture_thread)
        gesture_thread.daemon = True
        gesture_thread.start()

        # Start TUIO server
        try:
            self.log('Starting TUIO server')
            self.server.start()
        except SystemExit:
            self.stop()

    def bind(self, name, handler, *args, **kwargs):
        """
        Bind a handler to an event or gesture.
        """
        if name not in SUPPORTED_EVENTS:
            raise ValueError('Unsupported event name "%s".' % name)

        if name not in self.handlers:
            self.handlers[name] = []

        self.handlers[name].append((handler, args, kwargs))

    def unbind(self, name, handler=None):
        """
        Remove one or all handlers that are bound to an event name.
        """
        if name not in self.handlers:
            raise KeyError('No handlers are bound to event name "%s".' % name)

        if handler:
            self.handlers[name] = [h for h in self.handlers[name]
                                   if h[0] != handler]
        else:
            del self.handlers[name]

    def trigger(self, event):
        """
        Call all handlers bound to the name of the triggered event.
        """
        if event.__class__._name in self.handlers:
            h = self.handlers[event.__class__._name]
            self.log('Event triggered: "%s" (%d handlers)' % (event, len(h)),
                     1 + int(isinstance(event, Event)))

            for handler, args, kwargs in h:
                handler(event, *args, **kwargs)


if __name__ == '__main__':
    def tap(event, tap_type):
        print 'tap:', tap_type

    server = MultitouchServer(verbose=1, tuio_verbose=0)
    server.bind('tap', tap, 0)
    server.bind('single_tap', tap, 1)
    server.bind('double_tap', tap, 2)

    # Add empty handlers sp that the events are actually triggered
    server.bind('rotate', lambda e: 0)
    server.bind('pinch', lambda e: 0)
    server.bind('pan', lambda e: 0)

    try:
        server.start()
    except KeyboardInterrupt:
        server.stop()
